#!/usr/bin/env python3

from pybest import context
from pybest.cc import RCCSD
from pybest.gbasis import (
    compute_eri,
    compute_kinetic,
    compute_nuclear,
    compute_nuclear_repulsion,
    compute_overlap,
    get_gobasis,
)
from pybest.ip_eom import RIPCCSD
from pybest.linalg import DenseLinalgFactory
from pybest.occ_model import AufbauOccModel
from pybest.wrappers import RHF

###############################################################################
## Set up molecule, define basis set ##########################################
###############################################################################
# get the XYZ file from PyBEST's test data directory
fn_xyz = context.get_fn("test/no.xyz")
obasis = get_gobasis("cc-pvdz", fn_xyz)

###############################################################################
## Define Occupation model, expansion coefficients and overlap ################
###############################################################################
lf = DenseLinalgFactory(obasis.nbasis)
orb_a = lf.create_orbital(obasis.nbasis)
olp = compute_overlap(obasis)
# we need to add 1 additional electron
# If we do not specify the number of frozen core orbitals (ncore),
# then ncore will be calculated automatically
occ_model = AufbauOccModel(obasis, charge=-1, ncore=0)

###############################################################################
## Construct Hamiltonian ######################################################
###############################################################################
kin = compute_kinetic(obasis)
ne = compute_nuclear(obasis)
eri = compute_eri(obasis)
external = compute_nuclear_repulsion(obasis)

###############################################################################
## Do a Hartree-Fock calculation ##############################################
###############################################################################
hf = RHF(lf, occ_model)
hf_output = hf(kin, ne, eri, external, olp, orb_a)

###############################################################################
## Do CCSD optimization #######################################################
###############################################################################
ccsd = RCCSD(lf, occ_model)
ccsd_output = ccsd(kin, ne, eri, hf_output)

###############################################################################
### Do RIP-CCSD calculation for 1 unpaired electron ###########################
###############################################################################
ip = RIPCCSD(lf, occ_model, alpha=1)
ip_output = ip(kin, ne, eri, ccsd_output, nroot=8, nhole=2)
