# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Fixtures for tests"""

import numpy as np
import pytest

from pybest.geminals import ROOpCCD
from pybest.wrappers.hf import RHF
from pybest.wrappers.tests.common import Molecule

############################>Dipole moment<############################
#
# Define fixtures for testing: water and helium
# All in debye unit
# [ (molecule name), (expected) ]
test_hf_dm = [
    (
        "h2o",
        {
            "dipole_hf": [0.04129661, 2.25353156, 0.00000000],
        },
    ),
    (
        "water",
        {"dipole_hf": [0.00000000, -1.98081762, 0.00000000]},
    ),
    (
        "he",
        {"dipole_hf": [0.00000000, 0.00000000, 0.00000000]},
    ),
]

test_opccd_dm = [
    (
        "h2o",
        ROOpCCD,
        {
            "dipole_pccd": [0.2797343, 1.9267571, 0.00000000],
            "dipole_lccsd": [4.61536004e-02, 2.12667889, 0.00000000],
        },
    ),
    (
        "he",
        ROOpCCD,
        {
            "dipole_pccd": [0.00000000, 0.00000000, 0.00000000],
            "dipole_lccsd": [0.00000000, 0.00000000, 0.00000000],
        },
    ),
]

ncore = [0, 2]


@pytest.fixture()
def dm1_data() -> dict:
    # Elements of pccd density matrix for H2O/cc-pvdz without frozen core
    pccd_no_fc = np.array(
        [
            [
                1.09203776e00,
                1.37460640e-01,
                -7.19834024e-02,
                3.93385239e-02,
                6.19728070e-17,
                -5.45567330e-03,
                3.70272272e-01,
                1.91231757e-02,
                7.85211188e-01,
                -6.03180447e-01,
                1.50321532e-16,
                2.31595752e-01,
                -2.44912513e-01,
                8.09752639e-17,
                -2.95077259e-02,
                8.18164380e-19,
                -1.12673683e-02,
                -6.31062707e-18,
                -7.12548933e-03,
                -2.02500059e-01,
                -6.67032002e-02,
                1.71971648e-02,
                -2.92111362e-02,
                -5.90855272e-18,
            ],
            [
                1.37460640e-01,
                2.02391268e-02,
                -7.38720142e-03,
                4.86429624e-03,
                4.98093286e-18,
                6.87217196e-04,
                -1.90910545e-03,
                -1.70999852e-02,
                1.22877758e-01,
                -6.55092259e-02,
                -4.19973170e-17,
                3.71395926e-02,
                -2.98423959e-02,
                -2.21042814e-17,
                -4.49235568e-03,
                6.62938263e-19,
                -1.04472262e-03,
                -7.02430768e-19,
                -9.76756587e-04,
                -6.67032002e-02,
                -1.29963743e-02,
                -8.31149901e-04,
                -5.73012059e-03,
                -2.83122287e-18,
            ],
            [
                -7.19834024e-02,
                -7.38720142e-03,
                5.99064670e-03,
                -2.17674662e-03,
                -3.77070682e-18,
                1.69127194e-03,
                -4.84591125e-02,
                -8.59099098e-03,
                -3.38019265e-02,
                5.47093516e-02,
                7.56560581e-18,
                -9.60766622e-03,
                2.09358117e-02,
                4.72506630e-18,
                1.32521417e-03,
                -9.80889603e-19,
                1.12272735e-03,
                9.36651082e-19,
                7.57655753e-04,
                -1.71971648e-02,
                8.31149900e-04,
                -3.37145361e-03,
                7.59305072e-04,
                8.44965294e-19,
            ],
            [
                3.93385239e-02,
                4.86429624e-03,
                -2.17674662e-03,
                2.99877289e-03,
                2.78531995e-19,
                -1.12238721e-03,
                3.65465105e-02,
                1.92708327e-02,
                4.31474610e-02,
                1.19856125e-02,
                -4.89998215e-17,
                1.25927381e-02,
                7.77941439e-03,
                -2.50757477e-17,
                -1.46718086e-03,
                1.27307974e-18,
                1.30887939e-04,
                -1.14367484e-19,
                5.00301710e-04,
                -2.92111362e-02,
                -5.73012058e-03,
                -7.59305072e-04,
                -9.38802510e-04,
                -2.21078058e-18,
            ],
            [
                6.19728070e-17,
                4.98093286e-18,
                -3.77070682e-18,
                2.78531995e-19,
                2.08817222e-03,
                -1.07688495e-16,
                1.32887441e-17,
                8.80421232e-19,
                2.93692613e-17,
                -5.80904897e-17,
                5.53084108e-02,
                1.05865312e-17,
                -2.50500446e-17,
                3.02594522e-02,
                -3.16444689e-19,
                -1.20584475e-03,
                -2.74110278e-18,
                3.03414032e-04,
                8.85553015e-19,
                3.39062045e-18,
                3.33435780e-19,
                -3.22395068e-18,
                -3.23050525e-19,
                1.75596076e-03,
            ],
            [
                -5.45567330e-03,
                6.87217196e-04,
                1.69127194e-03,
                -1.12238721e-03,
                -1.07688495e-16,
                2.00335330e00,
                -2.42130444e-03,
                -1.61509785e-02,
                8.07563477e-13,
                -7.92029346e-03,
                3.39951940e-16,
                5.23655698e-13,
                -2.57482326e-03,
                6.24716832e-17,
                -1.11569345e-12,
                1.21501405e-17,
                -8.26116348e-04,
                -3.05494338e-17,
                -4.59329560e-04,
                -5.45567331e-03,
                6.87217194e-04,
                -1.69127194e-03,
                -1.12238721e-03,
                -3.33035782e-17,
            ],
            [
                3.70272272e-01,
                -1.90910545e-03,
                -4.84591125e-02,
                3.65465105e-02,
                1.32887441e-17,
                -2.42130444e-03,
                1.71554215e00,
                8.28579727e-01,
                1.48220821e-12,
                2.51454839e-01,
                -2.40037722e-16,
                3.91007117e-12,
                2.30092042e-01,
                -1.30135658e-17,
                -6.22069147e-12,
                3.11933027e-17,
                -3.14308437e-03,
                -1.70720983e-17,
                4.74361946e-03,
                3.70272272e-01,
                -1.90910547e-03,
                4.84591125e-02,
                3.65465105e-02,
                -1.18268787e-17,
            ],
            [
                1.91231757e-02,
                -1.70999852e-02,
                -8.59099098e-03,
                1.92708327e-02,
                8.80421232e-19,
                -1.61509785e-02,
                8.28579727e-01,
                5.14709048e-01,
                4.58826881e-12,
                4.48771574e-01,
                2.05435444e-16,
                2.33666051e-12,
                2.66746180e-01,
                1.00795802e-16,
                -4.74523985e-13,
                2.26942891e-19,
                6.36149115e-03,
                -7.80512313e-18,
                9.47599310e-03,
                1.91231757e-02,
                -1.70999852e-02,
                8.59099098e-03,
                1.92708327e-02,
                2.12285478e-18,
            ],
            [
                7.85211188e-01,
                1.22877758e-01,
                -3.38019265e-02,
                4.31474610e-02,
                2.93692613e-17,
                8.07563477e-13,
                1.48220821e-12,
                4.58826881e-12,
                1.03213542e00,
                6.19325741e-12,
                -2.57538784e-16,
                2.96138440e-01,
                1.06448881e-11,
                -1.25403143e-16,
                -3.48372692e-02,
                -1.04694112e-18,
                1.46839498e-13,
                1.34214220e-17,
                -8.14696130e-12,
                -7.85211188e-01,
                -1.22877758e-01,
                -3.38019265e-02,
                -4.31474610e-02,
                -2.60340838e-17,
            ],
            [
                -6.03180447e-01,
                -6.55092259e-02,
                5.47093516e-02,
                1.19856125e-02,
                -5.80904897e-17,
                -7.92029346e-03,
                2.51454839e-01,
                4.48771574e-01,
                6.19325741e-12,
                1.33045078e00,
                -3.53873013e-16,
                -2.58195946e-12,
                5.99600688e-01,
                -7.01705904e-17,
                5.37233414e-12,
                2.06093525e-17,
                2.21984427e-02,
                -1.18372762e-18,
                2.12847466e-02,
                -6.03180447e-01,
                -6.55092258e-02,
                -5.47093516e-02,
                1.19856125e-02,
                -3.67443140e-18,
            ],
            [
                1.50321532e-16,
                -4.19973170e-17,
                7.56560581e-18,
                -4.89998215e-17,
                5.53084108e-02,
                3.39951940e-16,
                -2.40037722e-16,
                2.05435444e-16,
                -2.57538784e-16,
                -3.53873013e-16,
                1.69189456e00,
                1.51777451e-17,
                -1.09619969e-16,
                8.96467791e-01,
                3.04093173e-17,
                -3.27212056e-02,
                -1.25710296e-16,
                6.28679307e-13,
                2.83352221e-17,
                -1.16456758e-16,
                4.28657897e-17,
                -1.64898186e-16,
                2.64274575e-17,
                5.53084108e-02,
            ],
            [
                2.31595752e-01,
                3.71395926e-02,
                -9.60766622e-03,
                1.25927381e-02,
                1.05865312e-17,
                5.23655698e-13,
                3.91007117e-12,
                2.33666051e-12,
                2.96138440e-01,
                -2.58195946e-12,
                1.51777451e-17,
                8.77520351e-02,
                1.24042213e-12,
                1.00154920e-17,
                -9.79301464e-03,
                -1.93929396e-18,
                -4.82277918e-13,
                1.72864535e-18,
                -6.72593412e-13,
                -2.31595752e-01,
                -3.71395926e-02,
                -9.60766622e-03,
                -1.25927381e-02,
                -4.02768756e-18,
            ],
            [
                -2.44912513e-01,
                -2.98423959e-02,
                2.09358117e-02,
                7.77941439e-03,
                -2.50500446e-17,
                -2.57482326e-03,
                2.30092042e-01,
                2.66746180e-01,
                1.06448881e-11,
                5.99600688e-01,
                -1.09619969e-16,
                1.24042213e-12,
                2.90032971e-01,
                -1.95557535e-17,
                -5.17825442e-14,
                8.43355199e-19,
                8.98699722e-03,
                -2.00926847e-18,
                1.01916078e-02,
                -2.44912513e-01,
                -2.98423959e-02,
                -2.09358117e-02,
                7.77941439e-03,
                -3.82223107e-18,
            ],
            [
                8.09752639e-17,
                -2.21042814e-17,
                4.72506630e-18,
                -2.50757477e-17,
                3.02594522e-02,
                6.24716832e-17,
                -1.30135658e-17,
                1.00795802e-16,
                -1.25403143e-16,
                -7.01705904e-17,
                8.96467791e-01,
                1.00154920e-17,
                -1.95557535e-17,
                5.00521716e-01,
                1.47655786e-17,
                -1.75107415e-02,
                -3.19241226e-17,
                1.85422658e-13,
                2.26342424e-17,
                -8.99075363e-17,
                1.78849023e-17,
                -8.62534521e-17,
                1.33794382e-17,
                3.02594522e-02,
            ],
            [
                -2.95077259e-02,
                -4.49235568e-03,
                1.32521417e-03,
                -1.46718086e-03,
                -3.16444689e-19,
                -1.11569345e-12,
                -6.22069147e-12,
                -4.74523985e-13,
                -3.48372692e-02,
                5.37233414e-12,
                3.04093173e-17,
                -9.79301464e-03,
                -5.17825442e-14,
                1.47655786e-17,
                2.00138051e-03,
                -4.78675340e-19,
                8.28720526e-13,
                1.11903037e-19,
                4.26917889e-13,
                2.95077259e-02,
                4.49235568e-03,
                1.32521417e-03,
                1.46718086e-03,
                1.53400402e-18,
            ],
            [
                8.18164380e-19,
                6.62938263e-19,
                -9.80889603e-19,
                1.27307974e-18,
                -1.20584475e-03,
                1.21501405e-17,
                3.11933027e-17,
                2.26942891e-19,
                -1.04694112e-18,
                2.06093525e-17,
                -3.27212056e-02,
                -1.93929396e-18,
                8.43355199e-19,
                -1.75107415e-02,
                -4.78675340e-19,
                1.92069489e-03,
                1.61281871e-18,
                1.39691640e-13,
                -9.16097445e-19,
                1.45460250e-17,
                1.99017342e-19,
                4.33143054e-18,
                2.64652958e-19,
                -1.20584475e-03,
            ],
            [
                -1.12673683e-02,
                -1.04472262e-03,
                1.12272735e-03,
                1.30887939e-04,
                -2.74110278e-18,
                -8.26116348e-04,
                -3.14308437e-03,
                6.36149115e-03,
                1.46839498e-13,
                2.21984427e-02,
                -1.25710296e-16,
                -4.82277918e-13,
                8.98699722e-03,
                -3.19241226e-17,
                8.28720526e-13,
                1.61281871e-18,
                2.14444883e-03,
                -1.06248476e-21,
                7.51461607e-04,
                -1.12673683e-02,
                -1.04472262e-03,
                -1.12272735e-03,
                1.30887939e-04,
                -1.84522187e-18,
            ],
            [
                -6.31062707e-18,
                -7.02430768e-19,
                9.36651082e-19,
                -1.14367484e-19,
                3.03414032e-04,
                -3.05494338e-17,
                -1.70720983e-17,
                -7.80512313e-18,
                1.34214220e-17,
                -1.18372762e-18,
                6.28679307e-13,
                1.72864535e-18,
                -2.00926847e-18,
                1.85422658e-13,
                1.11903037e-19,
                1.39691640e-13,
                -1.06248476e-21,
                1.12545931e-03,
                -3.42534535e-20,
                -1.87246655e-18,
                6.32203112e-19,
                -1.83165923e-19,
                -5.69284263e-19,
                -3.03414032e-04,
            ],
            [
                -7.12548933e-03,
                -9.76756587e-04,
                7.57655753e-04,
                5.00301710e-04,
                8.85553015e-19,
                -4.59329560e-04,
                4.74361946e-03,
                9.47599310e-03,
                -8.14696130e-12,
                2.12847466e-02,
                2.83352221e-17,
                -6.72593412e-13,
                1.01916078e-02,
                2.26342424e-17,
                4.26917889e-13,
                -9.16097445e-19,
                7.51461607e-04,
                -3.42534535e-20,
                1.52130145e-03,
                -7.12548934e-03,
                -9.76756588e-04,
                -7.57655754e-04,
                5.00301710e-04,
                1.61525881e-18,
            ],
            [
                -2.02500059e-01,
                -6.67032002e-02,
                -1.71971648e-02,
                -2.92111362e-02,
                3.39062045e-18,
                -5.45567331e-03,
                3.70272272e-01,
                1.91231757e-02,
                -7.85211188e-01,
                -6.03180447e-01,
                -1.16456758e-16,
                -2.31595752e-01,
                -2.44912513e-01,
                -8.99075363e-17,
                2.95077259e-02,
                1.45460250e-17,
                -1.12673683e-02,
                -1.87246655e-18,
                -7.12548934e-03,
                1.09203776e00,
                1.37460640e-01,
                7.19834024e-02,
                3.93385239e-02,
                3.59205147e-18,
            ],
            [
                -6.67032002e-02,
                -1.29963743e-02,
                8.31149900e-04,
                -5.73012058e-03,
                3.33435780e-19,
                6.87217194e-04,
                -1.90910547e-03,
                -1.70999852e-02,
                -1.22877758e-01,
                -6.55092258e-02,
                4.28657897e-17,
                -3.71395926e-02,
                -2.98423959e-02,
                1.78849023e-17,
                4.49235568e-03,
                1.99017342e-19,
                -1.04472262e-03,
                6.32203112e-19,
                -9.76756588e-04,
                1.37460640e-01,
                2.02391268e-02,
                7.38720142e-03,
                4.86429624e-03,
                2.70477273e-18,
            ],
            [
                1.71971648e-02,
                -8.31149901e-04,
                -3.37145361e-03,
                -7.59305072e-04,
                -3.22395068e-18,
                -1.69127194e-03,
                4.84591125e-02,
                8.59099098e-03,
                -3.38019265e-02,
                -5.47093516e-02,
                -1.64898186e-16,
                -9.60766622e-03,
                -2.09358117e-02,
                -8.62534521e-17,
                1.32521417e-03,
                4.33143054e-18,
                -1.12272735e-03,
                -1.83165923e-19,
                -7.57655754e-04,
                7.19834024e-02,
                7.38720142e-03,
                5.99064670e-03,
                2.17674662e-03,
                -5.32507584e-18,
            ],
            [
                -2.92111362e-02,
                -5.73012059e-03,
                7.59305072e-04,
                -9.38802510e-04,
                -3.23050525e-19,
                -1.12238721e-03,
                3.65465105e-02,
                1.92708327e-02,
                -4.31474610e-02,
                1.19856125e-02,
                2.64274575e-17,
                -1.25927381e-02,
                7.77941439e-03,
                1.33794382e-17,
                1.46718086e-03,
                2.64652958e-19,
                1.30887939e-04,
                -5.69284263e-19,
                5.00301710e-04,
                3.93385239e-02,
                4.86429624e-03,
                2.17674662e-03,
                2.99877289e-03,
                1.53028120e-18,
            ],
            [
                -5.90855272e-18,
                -2.83122287e-18,
                8.44965294e-19,
                -2.21078058e-18,
                1.75596076e-03,
                -3.33035782e-17,
                -1.18268787e-17,
                2.12285478e-18,
                -2.60340838e-17,
                -3.67443140e-18,
                5.53084108e-02,
                -4.02768756e-18,
                -3.82223107e-18,
                3.02594522e-02,
                1.53400402e-18,
                -1.20584475e-03,
                -1.84522187e-18,
                -3.03414032e-04,
                1.61525881e-18,
                3.59205147e-18,
                2.70477273e-18,
                -5.32507584e-18,
                1.53028120e-18,
                2.08817222e-03,
            ],
        ]
    )

    # Elements of pccd density matrix for H2O/cc-pvdz with frozen core

    pccd_fc = np.array(
        [
            [
                1.08935245e00,
                1.35756036e-01,
                -7.14884771e-02,
                4.07278909e-02,
                5.16786016e-17,
                -5.05265268e-03,
                3.72553851e-01,
                1.98129737e-02,
                7.86832324e-01,
                -6.02461125e-01,
                1.43351061e-16,
                2.31444571e-01,
                -2.44938738e-01,
                4.49718106e-17,
                -2.94188081e-02,
                -9.16088221e-18,
                -1.05338634e-02,
                4.86811589e-18,
                -6.93124702e-03,
                -2.08356656e-01,
                -6.64288754e-02,
                1.75168923e-02,
                -2.90325503e-02,
                -1.09803869e-17,
            ],
            [
                1.35756036e-01,
                2.02294834e-02,
                -7.29547356e-03,
                4.70545753e-03,
                4.31510095e-18,
                9.93424178e-04,
                -4.46278863e-03,
                -2.00696476e-02,
                1.20614904e-01,
                -6.78090612e-02,
                -5.31761186e-17,
                3.66142439e-02,
                -3.16401258e-02,
                -3.12901844e-17,
                -4.35365874e-03,
                6.50186936e-19,
                -1.17382024e-03,
                8.78151018e-20,
                -1.16842905e-03,
                -6.64288754e-02,
                -1.26773839e-02,
                -6.77486078e-04,
                -5.88124128e-03,
                -3.83987101e-18,
            ],
            [
                -7.14884771e-02,
                -7.29547356e-03,
                6.04920917e-03,
                -2.14016509e-03,
                -2.74692989e-18,
                1.59712933e-03,
                -5.01858593e-02,
                -9.04087654e-03,
                -3.33157821e-02,
                5.53535218e-02,
                -4.55833293e-19,
                -9.48057945e-03,
                2.17088618e-02,
                3.63205451e-18,
                1.29719825e-03,
                3.83016578e-19,
                1.17829006e-03,
                -4.10461262e-19,
                7.94345841e-04,
                -1.75168923e-02,
                6.77486078e-04,
                -3.58274893e-03,
                8.06143425e-04,
                9.69534984e-19,
            ],
            [
                4.07278909e-02,
                4.70545753e-03,
                -2.14016509e-03,
                3.27230108e-03,
                -3.10770331e-19,
                -1.39539578e-03,
                3.71498698e-02,
                2.08785095e-02,
                4.42002497e-02,
                1.24915819e-02,
                -4.76431712e-17,
                1.28084724e-02,
                8.71713340e-03,
                -2.28878619e-17,
                -1.53198745e-03,
                3.72649232e-19,
                3.09134570e-04,
                9.69725427e-20,
                6.81118790e-04,
                -2.90325503e-02,
                -5.88124128e-03,
                -8.06143425e-04,
                -7.88933017e-04,
                -2.18349498e-18,
            ],
            [
                5.16786016e-17,
                4.31510095e-18,
                -2.74692989e-18,
                -3.10770331e-19,
                2.09542936e-03,
                1.83956653e-16,
                -9.98911504e-18,
                -7.21789340e-18,
                2.60575564e-17,
                -5.66834904e-17,
                5.55216741e-02,
                1.03741309e-17,
                -2.23374238e-17,
                3.03542642e-02,
                -2.87925755e-19,
                -1.25281006e-03,
                -3.18390362e-18,
                2.91757279e-04,
                6.91196421e-19,
                -1.04589504e-17,
                -6.51731381e-19,
                -4.52408908e-18,
                -8.38854790e-19,
                1.77465463e-03,
            ],
            [
                -5.05265268e-03,
                9.93424178e-04,
                1.59712933e-03,
                -1.39539578e-03,
                1.83956653e-16,
                2.00215413e00,
                -9.50546138e-03,
                -1.43158531e-02,
                -6.46280695e-16,
                -4.13478358e-03,
                2.59769721e-16,
                -1.74770018e-16,
                -1.61033433e-03,
                -3.84475407e-17,
                1.31056097e-16,
                4.67191102e-17,
                -3.60243513e-04,
                1.39488083e-17,
                -5.79546046e-05,
                -5.05265268e-03,
                9.93424178e-04,
                -1.59712933e-03,
                -1.39539578e-03,
                8.72480278e-17,
            ],
            [
                3.72553851e-01,
                -4.46278863e-03,
                -5.01858593e-02,
                3.71498698e-02,
                -9.98911504e-18,
                -9.50546138e-03,
                1.68732304e00,
                8.48481164e-01,
                -3.08860868e-15,
                2.72111315e-01,
                -1.12413114e-16,
                1.03718644e-15,
                2.27545165e-01,
                -2.61676351e-17,
                -2.99038592e-15,
                -7.50760813e-18,
                8.15353106e-04,
                9.13199402e-18,
                8.01813143e-03,
                3.72553851e-01,
                -4.46278863e-03,
                5.01858593e-02,
                3.71498698e-02,
                -7.25792912e-18,
            ],
            [
                1.98129737e-02,
                -2.00696476e-02,
                -9.04087654e-03,
                2.08785095e-02,
                -7.21789340e-18,
                -1.43158531e-02,
                8.48481164e-01,
                5.17836862e-01,
                -2.29342781e-15,
                4.42451076e-01,
                1.61702784e-16,
                -2.83543864e-16,
                2.61713059e-01,
                1.25048420e-16,
                2.54747768e-16,
                -8.32590010e-18,
                6.25926089e-03,
                3.38284191e-18,
                9.27436793e-03,
                1.98129737e-02,
                -2.00696476e-02,
                9.04087654e-03,
                2.08785095e-02,
                4.95579903e-18,
            ],
            [
                7.86832324e-01,
                1.20614904e-01,
                -3.33157821e-02,
                4.42002497e-02,
                2.60575564e-17,
                -6.46280695e-16,
                -3.08860868e-15,
                -2.29342781e-15,
                1.03106596e00,
                5.02552435e-15,
                -2.53100553e-16,
                2.94799861e-01,
                -6.15096995e-15,
                -1.24603681e-16,
                -3.54377798e-02,
                -1.08760755e-18,
                1.53816836e-15,
                -6.59607908e-19,
                -5.06496628e-15,
                -7.86832324e-01,
                -1.20614904e-01,
                -3.33157821e-02,
                -4.42002497e-02,
                -2.25778800e-17,
            ],
            [
                -6.02461125e-01,
                -6.78090612e-02,
                5.53535218e-02,
                1.24915819e-02,
                -5.66834904e-17,
                -4.13478358e-03,
                2.72111315e-01,
                4.42451076e-01,
                5.02552435e-15,
                1.32508579e00,
                -3.73072286e-16,
                3.11304980e-15,
                5.94925035e-01,
                -9.35132334e-17,
                -5.67280282e-15,
                5.11751786e-18,
                2.01509244e-02,
                -5.00332983e-18,
                1.83647060e-02,
                -6.02461125e-01,
                -6.78090612e-02,
                -5.53535218e-02,
                1.24915819e-02,
                -5.18252914e-18,
            ],
            [
                1.43351061e-16,
                -5.31761186e-17,
                -4.55833293e-19,
                -4.76431712e-17,
                5.55216741e-02,
                2.59769721e-16,
                -1.12413114e-16,
                1.61702784e-16,
                -2.53100553e-16,
                -3.73072286e-16,
                1.69229453e00,
                1.55403462e-17,
                -1.33248789e-16,
                8.96338867e-01,
                3.02286337e-17,
                -3.29470826e-02,
                -8.35844353e-17,
                -1.46574866e-16,
                3.36849718e-17,
                -1.22964820e-16,
                3.53660266e-17,
                -1.56850915e-16,
                2.73914580e-17,
                5.55216741e-02,
            ],
            [
                2.31444571e-01,
                3.66142439e-02,
                -9.48057945e-03,
                1.28084724e-02,
                1.03741309e-17,
                -1.74770018e-16,
                1.03718644e-15,
                -2.83543864e-16,
                2.94799861e-01,
                3.11304980e-15,
                1.55403462e-17,
                8.73581960e-02,
                -5.04631592e-16,
                9.85754442e-18,
                -9.71845464e-03,
                -1.97481988e-18,
                7.41382503e-16,
                1.80866439e-19,
                -1.42650022e-15,
                -2.31444571e-01,
                -3.66142439e-02,
                -9.48057945e-03,
                -1.28084724e-02,
                -3.83761689e-18,
            ],
            [
                -2.44938738e-01,
                -3.16401258e-02,
                2.17088618e-02,
                8.71713340e-03,
                -2.23374238e-17,
                -1.61033433e-03,
                2.27545165e-01,
                2.61713059e-01,
                -6.15096995e-15,
                5.94925035e-01,
                -1.33248789e-16,
                -5.04631592e-16,
                2.91165528e-01,
                -2.46074112e-17,
                4.01665394e-16,
                9.70060291e-19,
                9.17342188e-03,
                -1.71426565e-18,
                1.01347057e-02,
                -2.44938738e-01,
                -3.16401258e-02,
                -2.17088618e-02,
                8.71713340e-03,
                -1.23999260e-18,
            ],
            [
                4.49718106e-17,
                -3.12901844e-17,
                3.63205451e-18,
                -2.28878619e-17,
                3.03542642e-02,
                -3.84475407e-17,
                -2.61676351e-17,
                1.25048420e-16,
                -1.24603681e-16,
                -9.35132334e-17,
                8.96338867e-01,
                9.85754442e-18,
                -2.46074112e-17,
                5.00137755e-01,
                1.48229344e-17,
                -1.74047194e-02,
                -4.08012911e-17,
                -3.10449365e-16,
                1.72247833e-17,
                -1.22611346e-16,
                1.08548805e-17,
                -8.49767466e-17,
                1.55357668e-17,
                3.03542642e-02,
            ],
            [
                -2.94188081e-02,
                -4.35365874e-03,
                1.29719825e-03,
                -1.53198745e-03,
                -2.87925755e-19,
                1.31056097e-16,
                -2.99038592e-15,
                2.54747768e-16,
                -3.54377798e-02,
                -5.67280282e-15,
                3.02286337e-17,
                -9.71845464e-03,
                4.01665394e-16,
                1.48229344e-17,
                1.96446859e-03,
                -4.61218394e-19,
                -3.62450659e-17,
                -6.31607863e-20,
                3.32567619e-16,
                2.94188081e-02,
                4.35365874e-03,
                1.29719825e-03,
                1.53198745e-03,
                1.51196921e-18,
            ],
            [
                -9.16088221e-18,
                6.50186936e-19,
                3.83016578e-19,
                3.72649232e-19,
                -1.25281006e-03,
                4.67191102e-17,
                -7.50760813e-18,
                -8.32590010e-18,
                -1.08760755e-18,
                5.11751786e-18,
                -3.29470826e-02,
                -1.97481988e-18,
                9.70060291e-19,
                -1.74047194e-02,
                -4.61218394e-19,
                1.90547521e-03,
                1.56124169e-18,
                -3.26267661e-17,
                -6.47240151e-19,
                4.53611995e-18,
                5.49100843e-20,
                2.98611625e-18,
                -6.31648829e-19,
                -1.25281006e-03,
            ],
            [
                -1.05338634e-02,
                -1.17382024e-03,
                1.17829006e-03,
                3.09134570e-04,
                -3.18390362e-18,
                -3.60243513e-04,
                8.15353106e-04,
                6.25926089e-03,
                1.53816836e-15,
                2.01509244e-02,
                -8.35844353e-17,
                7.41382503e-16,
                9.17342188e-03,
                -4.08012911e-17,
                -3.62450659e-17,
                1.56124169e-18,
                1.76562556e-03,
                -1.35140446e-19,
                5.18755628e-04,
                -1.05338634e-02,
                -1.17382024e-03,
                -1.17829006e-03,
                3.09134570e-04,
                -2.62368815e-18,
            ],
            [
                4.86811589e-18,
                8.78151018e-20,
                -4.10461262e-19,
                9.69725427e-20,
                2.91757279e-04,
                1.39488083e-17,
                9.13199402e-18,
                3.38284191e-18,
                -6.59607908e-19,
                -5.00332983e-18,
                -1.46574866e-16,
                1.80866439e-19,
                -1.71426565e-18,
                -3.10449365e-16,
                -6.31607863e-20,
                -3.26267661e-17,
                -1.35140446e-19,
                1.02448741e-03,
                -7.99620614e-20,
                5.91370398e-18,
                6.03320735e-19,
                6.71949938e-19,
                2.08555240e-19,
                -2.91757279e-04,
            ],
            [
                -6.93124702e-03,
                -1.16842905e-03,
                7.94345841e-04,
                6.81118790e-04,
                6.91196421e-19,
                -5.79546046e-05,
                8.01813143e-03,
                9.27436793e-03,
                -5.06496628e-15,
                1.83647060e-02,
                3.36849718e-17,
                -1.42650022e-15,
                1.01347057e-02,
                1.72247833e-17,
                3.32567619e-16,
                -6.47240151e-19,
                5.18755628e-04,
                -7.99620614e-20,
                1.16109846e-03,
                -6.93124702e-03,
                -1.16842905e-03,
                -7.94345841e-04,
                6.81118790e-04,
                1.10456289e-18,
            ],
            [
                -2.08356656e-01,
                -6.64288754e-02,
                -1.75168923e-02,
                -2.90325503e-02,
                -1.04589504e-17,
                -5.05265268e-03,
                3.72553851e-01,
                1.98129737e-02,
                -7.86832324e-01,
                -6.02461125e-01,
                -1.22964820e-16,
                -2.31444571e-01,
                -2.44938738e-01,
                -1.22611346e-16,
                2.94188081e-02,
                4.53611995e-18,
                -1.05338634e-02,
                5.91370398e-18,
                -6.93124702e-03,
                1.08935245e00,
                1.35756036e-01,
                7.14884771e-02,
                4.07278909e-02,
                2.27743847e-18,
            ],
            [
                -6.64288754e-02,
                -1.26773839e-02,
                6.77486078e-04,
                -5.88124128e-03,
                -6.51731381e-19,
                9.93424178e-04,
                -4.46278863e-03,
                -2.00696476e-02,
                -1.20614904e-01,
                -6.78090612e-02,
                3.53660266e-17,
                -3.66142439e-02,
                -3.16401258e-02,
                1.08548805e-17,
                4.35365874e-03,
                5.49100843e-20,
                -1.17382024e-03,
                6.03320735e-19,
                -1.16842905e-03,
                1.35756036e-01,
                2.02294834e-02,
                7.29547356e-03,
                4.70545753e-03,
                2.43864508e-18,
            ],
            [
                1.75168923e-02,
                -6.77486078e-04,
                -3.58274893e-03,
                -8.06143425e-04,
                -4.52408908e-18,
                -1.59712933e-03,
                5.01858593e-02,
                9.04087654e-03,
                -3.33157821e-02,
                -5.53535218e-02,
                -1.56850915e-16,
                -9.48057945e-03,
                -2.17088618e-02,
                -8.49767466e-17,
                1.29719825e-03,
                2.98611625e-18,
                -1.17829006e-03,
                6.71949938e-19,
                -7.94345841e-04,
                7.14884771e-02,
                7.29547356e-03,
                6.04920917e-03,
                2.14016509e-03,
                -5.17605621e-18,
            ],
            [
                -2.90325503e-02,
                -5.88124128e-03,
                8.06143425e-04,
                -7.88933017e-04,
                -8.38854790e-19,
                -1.39539578e-03,
                3.71498698e-02,
                2.08785095e-02,
                -4.42002497e-02,
                1.24915819e-02,
                2.73914580e-17,
                -1.28084724e-02,
                8.71713340e-03,
                1.55357668e-17,
                1.53198745e-03,
                -6.31648829e-19,
                3.09134570e-04,
                2.08555240e-19,
                6.81118790e-04,
                4.07278909e-02,
                4.70545753e-03,
                2.14016509e-03,
                3.27230108e-03,
                1.45295065e-18,
            ],
            [
                -1.09803869e-17,
                -3.83987101e-18,
                9.69534984e-19,
                -2.18349498e-18,
                1.77465463e-03,
                8.72480278e-17,
                -7.25792912e-18,
                4.95579903e-18,
                -2.25778800e-17,
                -5.18252914e-18,
                5.55216741e-02,
                -3.83761689e-18,
                -1.23999260e-18,
                3.03542642e-02,
                1.51196921e-18,
                -1.25281006e-03,
                -2.62368815e-18,
                -2.91757279e-04,
                1.10456289e-18,
                2.27743847e-18,
                2.43864508e-18,
                -5.17605621e-18,
                1.45295065e-18,
                2.09542936e-03,
            ],
        ]
    )

    # Elements of pccd-lccsd density matrix for H2O/cc-pvdz without frozen core
    lccsd_no_fc = [
        [
            1.10685540e00,
            1.47961502e-01,
            -6.74228057e-02,
            3.55608771e-02,
            5.75634489e-17,
            -3.65416888e-03,
            3.63385280e-01,
            -1.36792370e-03,
            7.78676541e-01,
            -5.98615784e-01,
            8.85221204e-17,
            2.19319647e-01,
            -2.38196639e-01,
            1.87850377e-17,
            -2.82396796e-02,
            -8.61691319e-19,
            -1.16779730e-02,
            -2.87096355e-18,
            -7.83633235e-03,
            -1.84940582e-01,
            -6.78968066e-02,
            1.92285445e-02,
            -2.94581756e-02,
            -1.37577401e-17,
        ],
        [
            1.31733660e-01,
            2.10395545e-02,
            -6.75111574e-03,
            4.27424607e-03,
            2.27624422e-18,
            1.61732029e-03,
            -2.35295915e-03,
            -2.24218516e-02,
            1.24671219e-01,
            -7.16990446e-02,
            -8.27500597e-17,
            3.55693263e-02,
            -3.37135589e-02,
            -4.85529566e-17,
            -4.01239250e-03,
            8.47059777e-19,
            -1.26523418e-03,
            8.56453267e-19,
            -1.33698256e-03,
            -6.02140649e-02,
            -1.22851420e-02,
            -4.54754009e-04,
            -5.87090459e-03,
            -5.63242627e-18,
        ],
        [
            -7.27782748e-02,
            -7.89090313e-03,
            6.06157199e-03,
            -1.82315744e-03,
            -4.12348288e-18,
            1.65269762e-03,
            -4.77763420e-02,
            -6.97535637e-03,
            -3.05670166e-02,
            5.26293189e-02,
            -1.64773452e-18,
            -8.66397964e-03,
            1.99640215e-02,
            9.08406629e-19,
            1.23500472e-03,
            -1.19040398e-18,
            9.44379420e-04,
            1.03148966e-18,
            8.67296150e-04,
            -1.89847884e-02,
            5.26988447e-04,
            -3.51505294e-03,
            7.89087386e-04,
            5.75997796e-19,
        ],
        [
            3.90250445e-02,
            4.94742124e-03,
            -1.89259143e-03,
            3.56773214e-03,
            -1.96921009e-20,
            -1.53181096e-03,
            3.76744088e-02,
            2.19393801e-02,
            4.17571145e-02,
            1.55486628e-02,
            -4.59492024e-17,
            1.19519267e-02,
            1.06419825e-02,
            -2.00701632e-17,
            -1.38772553e-03,
            9.77085151e-19,
            1.93084518e-04,
            -7.05643763e-19,
            7.76386692e-04,
            -2.99618814e-02,
            -6.41963602e-03,
            -6.08844869e-04,
            -5.75999067e-04,
            -1.95545318e-18,
        ],
        [
            5.46775788e-17,
            2.75529021e-18,
            -3.27394206e-18,
            -2.10760510e-19,
            3.27668108e-03,
            -1.07987979e-16,
            1.60222916e-17,
            8.20615905e-18,
            2.22588529e-17,
            -5.81395394e-17,
            5.85481403e-02,
            9.22985050e-18,
            -1.81399279e-17,
            3.35493743e-02,
            -4.40158983e-20,
            -1.48443003e-03,
            -4.46934760e-18,
            3.82429169e-04,
            5.20884888e-19,
            -2.20870943e-18,
            -3.87446938e-19,
            -3.88057661e-18,
            -2.67987312e-20,
            2.17298750e-03,
        ],
        [
            -5.27307550e-03,
            9.75214768e-04,
            1.72020400e-03,
            -1.59639503e-03,
            -1.07929315e-16,
            2.00469985e00,
            3.37875445e-03,
            -1.84985580e-02,
            2.71189372e-12,
            -7.59017806e-03,
            3.58738473e-16,
            3.95800222e-13,
            -2.94629392e-03,
            5.83961424e-17,
            2.72428351e-13,
            9.33030144e-18,
            -1.21119879e-04,
            -3.08375343e-17,
            4.07158329e-04,
            -5.27307550e-03,
            9.75214768e-04,
            -1.72020400e-03,
            -1.59639503e-03,
            -3.35430889e-17,
        ],
        [
            3.66471908e-01,
            -1.68426344e-04,
            -4.79480673e-02,
            3.65178946e-02,
            1.56392318e-17,
            3.85272581e-03,
            1.74608631e00,
            8.21582194e-01,
            4.10149326e-12,
            2.46611674e-01,
            -1.25778234e-16,
            -1.37642051e-12,
            2.34862132e-01,
            2.23814968e-18,
            2.23147048e-12,
            1.63207474e-17,
            1.63461401e-03,
            -1.88963661e-17,
            1.12007700e-02,
            3.66471908e-01,
            -1.68426338e-04,
            4.79480673e-02,
            3.65178946e-02,
            -1.06299030e-17,
        ],
        [
            9.14721130e-03,
            -1.95663587e-02,
            -8.57303702e-03,
            2.22454624e-02,
            6.87154645e-18,
            -1.93198862e-02,
            8.26399865e-01,
            5.29714742e-01,
            -1.35537066e-11,
            4.42194700e-01,
            1.98865486e-16,
            -3.72942373e-12,
            2.68445056e-01,
            1.60318501e-16,
            9.52706837e-13,
            6.20980787e-18,
            5.34579249e-03,
            -7.01099209e-18,
            9.29231168e-03,
            9.14721133e-03,
            -1.95663587e-02,
            8.57303702e-03,
            2.22454624e-02,
            7.62269538e-18,
        ],
        [
            7.84481903e-01,
            1.32400337e-01,
            -2.84513182e-02,
            4.03818077e-02,
            2.51129952e-17,
            -2.89508587e-12,
            -3.80222148e-12,
            9.11159990e-12,
            1.06218407e00,
            1.11476069e-12,
            -2.54453274e-16,
            2.84431828e-01,
            -7.61681828e-12,
            -1.22956774e-16,
            -3.57132762e-02,
            -2.07191742e-18,
            4.80198663e-13,
            -7.12218925e-18,
            1.71469458e-12,
            -7.84481903e-01,
            -1.32400337e-01,
            -2.84513182e-02,
            -4.03818077e-02,
            -2.19637716e-17,
        ],
        [
            -6.01058385e-01,
            -7.28934796e-02,
            5.03703245e-02,
            1.69081841e-02,
            -6.79190170e-17,
            -9.38357146e-03,
            2.41217380e-01,
            4.46856739e-01,
            4.29461820e-12,
            1.33301432e00,
            -3.89663538e-16,
            4.53198620e-12,
            5.81385605e-01,
            -1.02451576e-16,
            -1.94380836e-12,
            -8.87783399e-18,
            1.94892536e-02,
            -6.11572580e-19,
            2.04062798e-02,
            -6.01058385e-01,
            -7.28934796e-02,
            -5.03703245e-02,
            1.69081841e-02,
            -8.92013918e-18,
        ],
        [
            1.45259143e-16,
            -6.44639191e-17,
            1.30905978e-18,
            -4.96716063e-17,
            5.85211057e-02,
            3.58128456e-16,
            -1.67947734e-16,
            1.52187679e-16,
            -2.57324771e-16,
            -3.54936090e-16,
            1.67895825e00,
            1.91871478e-17,
            -1.23159604e-16,
            8.85358850e-01,
            3.15023299e-17,
            -3.19543638e-02,
            -7.62060501e-17,
            -2.08291380e-13,
            4.69557420e-17,
            -1.25046290e-16,
            1.97806303e-17,
            -1.62381741e-16,
            2.48196695e-17,
            5.85211057e-02,
        ],
        [
            2.30764143e-01,
            4.04489802e-02,
            -8.34385130e-03,
            1.18319382e-02,
            1.04010305e-17,
            -1.88107430e-13,
            5.52239134e-12,
            1.24526567e-12,
            2.93416388e-01,
            -1.21259213e-11,
            2.01431739e-17,
            8.70601316e-02,
            -6.63169903e-12,
            1.19505029e-17,
            -9.12982987e-03,
            -2.14789829e-18,
            -3.50486275e-13,
            1.56602875e-19,
            -1.36664424e-13,
            -2.30764143e-01,
            -4.04489802e-02,
            -8.34385130e-03,
            -1.18319382e-02,
            -3.85987493e-18,
        ],
        [
            -2.41573623e-01,
            -3.44138415e-02,
            1.89380879e-02,
            1.14812201e-02,
            -2.24534108e-17,
            -3.34394535e-03,
            2.25099754e-01,
            2.64671305e-01,
            8.07011121e-12,
            5.84459769e-01,
            -1.52388130e-16,
            3.11460479e-12,
            2.84689036e-01,
            -5.07673808e-17,
            -2.02613664e-13,
            6.02432788e-18,
            8.70892160e-03,
            -1.69532027e-18,
            9.86138772e-03,
            -2.41573624e-01,
            -3.44138415e-02,
            -1.89380879e-02,
            1.14812201e-02,
            -1.73954560e-18,
        ],
        [
            5.05000957e-17,
            -3.63014607e-17,
            3.35888526e-18,
            -2.18787953e-17,
            3.36612225e-02,
            5.51519696e-17,
            -1.64943300e-17,
            1.55683574e-16,
            -1.22931093e-16,
            -1.03965331e-16,
            8.85893535e-01,
            1.13563182e-17,
            -4.07786761e-17,
            5.02913520e-01,
            1.49985377e-17,
            -1.77063884e-02,
            -5.12867817e-17,
            -2.15989524e-13,
            1.54160666e-17,
            -1.10520688e-16,
            5.64381905e-18,
            -8.28068712e-17,
            1.56950015e-17,
            3.36612225e-02,
        ],
        [
            -2.93767759e-02,
            -4.60277010e-03,
            1.19864583e-03,
            -1.34317726e-03,
            -3.34902750e-19,
            1.80623683e-13,
            1.92336006e-12,
            -7.50995528e-14,
            -3.63168321e-02,
            -2.91159367e-12,
            3.06883560e-17,
            -8.48916553e-03,
            -6.65244014e-13,
            1.44096285e-17,
            4.14006790e-03,
            -5.71457122e-19,
            -1.45568395e-13,
            -5.31972881e-19,
            6.14267161e-16,
            2.93767759e-02,
            4.60277011e-03,
            1.19864583e-03,
            1.34317726e-03,
            1.54552590e-18,
        ],
        [
            3.96361375e-18,
            1.95376955e-18,
            -1.25424062e-18,
            5.01297655e-19,
            -1.60569110e-03,
            1.00294078e-17,
            1.31486036e-17,
            -1.56758797e-18,
            -2.06907681e-18,
            -1.39252038e-17,
            -3.15744237e-02,
            -2.10495974e-18,
            -3.01706449e-18,
            -1.72586543e-02,
            -6.06375913e-19,
            3.37784237e-03,
            1.34063211e-18,
            5.16717975e-14,
            -9.63887941e-19,
            1.74349672e-17,
            1.25486281e-18,
            4.63446330e-18,
            -4.68697700e-19,
            -1.60569110e-03,
        ],
        [
            -1.15541866e-02,
            -1.09063618e-03,
            1.04662061e-03,
            2.63032282e-04,
            -3.42696322e-18,
            -3.37032381e-04,
            1.92482841e-03,
            6.72676519e-03,
            -1.15870879e-12,
            2.00151430e-02,
            -7.05438633e-17,
            -5.74273789e-14,
            8.02433932e-03,
            -3.38537999e-17,
            -9.77877174e-14,
            1.49899590e-18,
            2.84659732e-03,
            7.35887488e-20,
            3.12586037e-04,
            -1.15541866e-02,
            -1.09063618e-03,
            -1.04662061e-03,
            2.63032282e-04,
            -2.75015178e-18,
        ],
        [
            -8.06091913e-18,
            -1.41414689e-18,
            9.44710836e-19,
            -5.59431470e-19,
            5.61519498e-04,
            -3.08645346e-17,
            -1.89534913e-17,
            -7.09653401e-18,
            -5.87523667e-18,
            -5.21497931e-19,
            -6.54807435e-13,
            -8.84115043e-20,
            -1.80673092e-18,
            -3.33641011e-13,
            -2.56645478e-21,
            3.08614778e-13,
            8.10303448e-20,
            3.05526069e-03,
            -7.27852970e-20,
            8.91004453e-20,
            1.44266301e-18,
            -2.81172837e-19,
            -7.40051286e-20,
            -5.61519498e-04,
        ],
        [
            -6.79413872e-03,
            -1.14580377e-03,
            8.95140328e-04,
            9.47161594e-04,
            3.11683599e-19,
            2.78635710e-04,
            1.13190046e-02,
            9.53137241e-03,
            1.60309960e-12,
            1.98862520e-02,
            4.36425034e-17,
            4.80657270e-13,
            9.88357759e-03,
            1.51805291e-17,
            -1.49261511e-13,
            -6.55325021e-19,
            2.14344014e-04,
            -7.19436079e-20,
            2.92747102e-03,
            -6.79413873e-03,
            -1.14580377e-03,
            -8.95140328e-04,
            9.47161594e-04,
            7.49662212e-19,
        ],
        [
            -1.84940582e-01,
            -6.78968066e-02,
            -1.92285445e-02,
            -2.94581756e-02,
            -3.36684113e-18,
            -3.65416888e-03,
            3.63385280e-01,
            -1.36792371e-03,
            -7.78676541e-01,
            -5.98615784e-01,
            -1.83197478e-16,
            -2.19319647e-01,
            -2.38196639e-01,
            -1.37941692e-16,
            2.82396796e-02,
            1.32914511e-17,
            -1.16779730e-02,
            -5.65117959e-18,
            -7.83633235e-03,
            1.10685540e00,
            1.47961502e-01,
            6.74228057e-02,
            3.55608771e-02,
            -2.34424317e-18,
        ],
        [
            -6.02140649e-02,
            -1.22851420e-02,
            4.54754008e-04,
            -5.87090459e-03,
            -1.28769820e-18,
            1.61732029e-03,
            -2.35295914e-03,
            -2.24218516e-02,
            -1.24671219e-01,
            -7.16990446e-02,
            3.77608456e-18,
            -3.55693263e-02,
            -3.37135589e-02,
            -1.31161887e-18,
            4.01239250e-03,
            2.10882922e-19,
            -1.26523418e-03,
            -8.38773429e-19,
            -1.33698256e-03,
            1.31733660e-01,
            2.10395545e-02,
            6.75111574e-03,
            4.27424607e-03,
            6.42971740e-19,
        ],
        [
            1.89847884e-02,
            -5.26988448e-04,
            -3.51505294e-03,
            -7.89087386e-04,
            -3.03365957e-18,
            -1.65269762e-03,
            4.77763420e-02,
            6.97535637e-03,
            -3.05670166e-02,
            -5.26293189e-02,
            -1.56561773e-16,
            -8.66397964e-03,
            -1.99640215e-02,
            -7.80016208e-17,
            1.23500472e-03,
            4.56206062e-18,
            -9.44379420e-04,
            -2.84328737e-19,
            -8.67296150e-04,
            7.27782748e-02,
            7.89090313e-03,
            6.06157199e-03,
            1.82315744e-03,
            -4.89934546e-18,
        ],
        [
            -2.99618814e-02,
            -6.41963602e-03,
            6.08844869e-04,
            -5.75999067e-04,
            -1.26094862e-19,
            -1.53181096e-03,
            3.76744088e-02,
            2.19393801e-02,
            -4.17571145e-02,
            1.55486628e-02,
            2.74089652e-17,
            -1.19519267e-02,
            1.06419825e-02,
            1.66204299e-17,
            1.38772553e-03,
            6.52321061e-20,
            1.93084518e-04,
            7.02990768e-20,
            7.76386692e-04,
            3.90250445e-02,
            4.94742124e-03,
            1.89259143e-03,
            3.56773214e-03,
            1.42074499e-18,
        ],
        [
            -1.11502632e-17,
            -4.32049287e-18,
            1.18350531e-18,
            -1.93060465e-18,
            2.17298750e-03,
            -3.37210736e-17,
            -1.07157694e-17,
            8.82976428e-18,
            -1.95358800e-17,
            -5.42615890e-19,
            5.85481403e-02,
            -2.99461052e-18,
            1.24483905e-18,
            3.35493743e-02,
            1.32247451e-18,
            -1.48443003e-03,
            -3.85914421e-18,
            -3.82429169e-04,
            9.96458890e-19,
            -4.47630647e-18,
            8.31261016e-19,
            -5.71310273e-18,
            1.33599430e-18,
            3.27668108e-03,
        ],
    ]

    # Elements of pccd-lccsd density matrix for H2O/cc-pvdz with frozen core
    lccsd_fc = [
        [
            1.09164496e00,
            1.39117837e-01,
            -6.99027040e-02,
            3.61701941e-02,
            5.27229579e-17,
            -3.89462688e-03,
            3.70148125e-01,
            6.79883311e-03,
            7.82252075e-01,
            -6.03932040e-01,
            1.47704275e-16,
            2.23992386e-01,
            -2.47073170e-01,
            4.60171780e-17,
            -2.87746784e-02,
            -9.14460630e-18,
            -1.23052572e-02,
            4.80572868e-18,
            -7.18687310e-03,
            -1.82825857e-01,
            -6.21034004e-02,
            1.95647158e-02,
            -3.00460554e-02,
            -1.03928023e-17,
        ],
        [
            1.25162373e-01,
            1.81259125e-02,
            -7.04321033e-03,
            4.18891165e-03,
            3.91349225e-18,
            1.04793075e-03,
            -4.99514672e-03,
            -2.10149161e-02,
            1.16846597e-01,
            -6.92258663e-02,
            -5.19851620e-17,
            3.35266497e-02,
            -3.21740739e-02,
            -3.21035015e-17,
            -3.76223276e-03,
            6.83612931e-19,
            -1.03132924e-03,
            -1.51468166e-19,
            -1.11830271e-03,
            -5.44533618e-02,
            -9.61597179e-03,
            -5.33913963e-04,
            -5.85934154e-03,
            -3.75203657e-18,
        ],
        [
            -7.42064752e-02,
            -7.99132506e-03,
            5.98878248e-03,
            -1.99712643e-03,
            -2.60309513e-18,
            1.57743806e-03,
            -4.97714495e-02,
            -8.12476996e-03,
            -3.33347898e-02,
            5.41620151e-02,
            4.74261736e-19,
            -9.63776413e-03,
            2.14321739e-02,
            4.13879860e-18,
            1.34537162e-03,
            2.09153498e-19,
            1.02965293e-03,
            -1.26590317e-19,
            8.99943520e-04,
            -1.78017202e-02,
            8.46875523e-04,
            -3.58554111e-03,
            8.51918455e-04,
            9.73601224e-19,
        ],
        [
            3.91264596e-02,
            4.78786441e-03,
            -2.01642557e-03,
            3.27243895e-03,
            -7.13887258e-19,
            -1.24259781e-03,
            3.88881985e-02,
            2.16317976e-02,
            4.31553775e-02,
            1.46104139e-02,
            -4.74296832e-17,
            1.26425560e-02,
            9.36303961e-03,
            -2.26231215e-17,
            -1.38747837e-03,
            3.37933851e-19,
            8.74642656e-05,
            4.57525391e-20,
            7.93313652e-04,
            -3.04673008e-02,
            -6.31631081e-03,
            -6.77374849e-04,
            -5.91985709e-04,
            -2.26277288e-18,
        ],
        [
            5.19347135e-17,
            4.13791444e-18,
            -2.48028282e-18,
            -7.69815408e-19,
            2.90409560e-03,
            1.83870988e-16,
            -1.06428829e-17,
            -7.02330062e-18,
            2.54576720e-17,
            -5.88919249e-17,
            5.87256071e-02,
            1.00809845e-17,
            -2.20521121e-17,
            3.32108558e-02,
            -2.50037559e-19,
            -1.56587251e-03,
            -3.46172948e-18,
            4.92430952e-04,
            6.60976079e-19,
            -1.03921486e-17,
            -1.79495061e-19,
            -4.68183851e-18,
            -9.11097298e-19,
            2.13249692e-03,
        ],
        [
            -4.21274906e-03,
            1.02212699e-03,
            1.66100943e-03,
            -1.28939425e-03,
            1.83951737e-16,
            2.00216722e00,
            -9.73759295e-03,
            -1.48169102e-02,
            3.77775712e-16,
            -4.01599268e-03,
            2.59743676e-16,
            -1.29952375e-16,
            -1.69558243e-03,
            -3.78783508e-17,
            5.37647475e-16,
            4.67146066e-17,
            -1.71865207e-04,
            1.38132310e-17,
            8.33399126e-05,
            -4.21274906e-03,
            1.02212699e-03,
            -1.66100943e-03,
            -1.28939425e-03,
            8.73316901e-17,
        ],
        [
            3.70520459e-01,
            -7.15683343e-03,
            -5.03198951e-02,
            3.95877859e-02,
            -1.09850123e-17,
            -9.88457332e-03,
            1.68779875e00,
            8.51579257e-01,
            -4.89444325e-17,
            2.72434004e-01,
            -1.09506439e-16,
            -1.15030505e-15,
            2.25261663e-01,
            -2.34046939e-17,
            1.82695739e-16,
            -8.17176346e-18,
            2.03401130e-03,
            8.92740895e-18,
            8.97607528e-03,
            3.70520459e-01,
            -7.15683343e-03,
            5.03198951e-02,
            3.95877859e-02,
            -7.58499765e-18,
        ],
        [
            8.90459371e-03,
            -2.34905205e-02,
            -9.45075991e-03,
            2.28323671e-02,
            -8.13965223e-18,
            -1.51313528e-02,
            8.50364680e-01,
            5.27438661e-01,
            -2.42601996e-15,
            4.40987177e-01,
            1.58681948e-16,
            -6.46335714e-16,
            2.61182868e-01,
            1.15190813e-16,
            -1.02025145e-17,
            -8.22707776e-18,
            6.44373183e-03,
            3.66981005e-18,
            9.16502630e-03,
            8.90459371e-03,
            -2.34905205e-02,
            9.45075991e-03,
            2.28323671e-02,
            4.41695789e-18,
        ],
        [
            7.91851601e-01,
            1.24075203e-01,
            -3.07917831e-02,
            4.24323456e-02,
            2.62725374e-17,
            -1.87331635e-15,
            -3.26372489e-15,
            1.50507852e-14,
            1.06220153e00,
            -2.33262438e-15,
            -2.51591160e-16,
            2.92402057e-01,
            3.44415956e-15,
            -1.21802036e-16,
            -3.38890519e-02,
            -2.30745513e-18,
            9.54201947e-17,
            -1.49575492e-18,
            5.97084706e-16,
            -7.91851601e-01,
            -1.24075203e-01,
            -3.07917831e-02,
            -4.24323456e-02,
            -2.30561225e-17,
        ],
        [
            -6.04749708e-01,
            -7.38064474e-02,
            5.28942951e-02,
            1.66253546e-02,
            -6.01893691e-17,
            -4.54423280e-03,
            2.72798464e-01,
            4.44977131e-01,
            9.48824431e-15,
            1.32229503e00,
            -3.67507132e-16,
            -3.77419508e-15,
            5.83061667e-01,
            -8.58219298e-17,
            6.92629567e-16,
            3.61495621e-18,
            1.89813162e-02,
            -4.79307715e-18,
            2.04174679e-02,
            -6.04749708e-01,
            -7.38064474e-02,
            -5.28942951e-02,
            1.66253546e-02,
            -5.85435457e-18,
        ],
        [
            1.47374036e-16,
            -5.29097508e-17,
            7.74165422e-19,
            -4.79965286e-17,
            5.89921125e-02,
            2.58649572e-16,
            -1.13890375e-16,
            1.65433317e-16,
            -2.55270254e-16,
            -3.68674269e-16,
            1.68423264e00,
            1.97058922e-17,
            -1.26337718e-16,
            8.89291592e-01,
            3.11376358e-17,
            -3.14941563e-02,
            -8.31760514e-17,
            1.85012362e-16,
            4.11921157e-17,
            -1.25979470e-16,
            3.60775036e-17,
            -1.61558690e-16,
            2.69567140e-17,
            5.89921125e-02,
        ],
        [
            2.36671833e-01,
            3.84325900e-02,
            -9.19466516e-03,
            1.27142226e-02,
            1.06977826e-17,
            -4.63006214e-16,
            2.26469457e-15,
            6.43462606e-15,
            3.01405521e-01,
            7.38748153e-15,
            2.07034017e-17,
            8.98896968e-02,
            2.01543700e-15,
            1.25187046e-17,
            -8.95548736e-03,
            -2.16271834e-18,
            -3.38172757e-17,
            3.58090603e-19,
            1.38765287e-16,
            -2.36671833e-01,
            -3.84325900e-02,
            -9.19466516e-03,
            -1.27142226e-02,
            -3.96990958e-18,
        ],
        [
            -2.47654893e-01,
            -3.49160878e-02,
            2.03437998e-02,
            1.07225460e-02,
            -2.34092143e-17,
            -1.58952590e-03,
            2.28206666e-01,
            2.62564082e-01,
            -9.82046133e-15,
            5.85205957e-01,
            -1.23195065e-16,
            -3.59194255e-15,
            2.81587500e-01,
            -2.37008545e-17,
            5.51740450e-16,
            1.04739491e-18,
            8.23828946e-03,
            -1.44433713e-18,
            1.02040574e-02,
            -2.47654893e-01,
            -3.49160878e-02,
            -2.03437998e-02,
            1.07225460e-02,
            -1.99724901e-18,
        ],
        [
            4.67608480e-17,
            -3.16759224e-17,
            3.84813430e-18,
            -2.33080572e-17,
            3.34890632e-02,
            -3.84624624e-17,
            -2.48749109e-17,
            1.20524121e-16,
            -1.22731252e-16,
            -8.67813291e-17,
            8.89783700e-01,
            1.18303198e-17,
            -2.56834367e-17,
            4.98460716e-01,
            1.51314758e-17,
            -1.84216078e-02,
            -4.12517381e-17,
            -1.49182235e-17,
            1.84429584e-17,
            -1.14532443e-16,
            1.30300279e-17,
            -8.45072159e-17,
            1.47966170e-17,
            3.34890632e-02,
        ],
        [
            -2.99501687e-02,
            -4.24718544e-03,
            1.30486803e-03,
            -1.37115816e-03,
            -3.46086382e-19,
            6.96272256e-16,
            8.31952017e-16,
            -4.56435191e-16,
            -3.46885066e-02,
            6.60224526e-16,
            3.03738064e-17,
            -8.34189695e-03,
            -3.09328435e-16,
            1.45494921e-17,
            3.25744140e-03,
            -4.96720097e-19,
            1.25616669e-16,
            -2.83423882e-19,
            -9.78198996e-17,
            2.99501687e-02,
            4.24718544e-03,
            1.30486803e-03,
            1.37115816e-03,
            1.62430001e-18,
        ],
        [
            -8.89826039e-18,
            7.83897021e-19,
            2.18781722e-19,
            3.67491224e-19,
            -1.70711356e-03,
            4.67559355e-17,
            -8.19102178e-18,
            -8.35333068e-18,
            -2.25095956e-18,
            3.40939603e-18,
            -3.12075834e-02,
            -2.12845759e-18,
            9.18998283e-19,
            -1.79395196e-02,
            -5.40782082e-19,
            2.88523986e-03,
            1.48215755e-18,
            1.92613697e-16,
            -5.57126499e-19,
            4.60083075e-18,
            -8.16254393e-20,
            2.86864171e-18,
            -5.33672439e-19,
            -1.70711356e-03,
        ],
        [
            -1.30058430e-02,
            -1.19685611e-03,
            1.19562954e-03,
            2.22089404e-04,
            -3.44635788e-18,
            -2.84354760e-04,
            1.77888302e-03,
            7.29384448e-03,
            -9.53175011e-16,
            1.94860953e-02,
            -8.22690236e-17,
            -4.40210672e-16,
            7.77014432e-03,
            -3.99428761e-17,
            -9.83547084e-18,
            1.44098454e-18,
            2.48569597e-03,
            -1.02902502e-19,
            3.80612313e-04,
            -1.30058430e-02,
            -1.19685611e-03,
            -1.19562954e-03,
            2.22089404e-04,
            -2.86552717e-18,
        ],
        [
            4.88197727e-18,
            -1.45795616e-19,
            -2.01745724e-19,
            3.76600082e-20,
            6.71235511e-04,
            1.38273732e-17,
            9.00635647e-18,
            3.44297070e-18,
            -1.41589150e-18,
            -4.72992562e-18,
            3.42069660e-16,
            4.16891681e-19,
            -1.48692904e-18,
            2.79342700e-16,
            -2.45184790e-19,
            -2.69433727e-16,
            -9.55578024e-20,
            2.38464382e-03,
            -2.62024023e-21,
            6.00347200e-18,
            9.38484309e-19,
            7.43110593e-19,
            2.63526594e-19,
            -6.71235511e-04,
        ],
        [
            -7.28733461e-03,
            -1.28765335e-03,
            9.33859117e-04,
            9.51091881e-04,
            4.94476983e-19,
            6.91668512e-05,
            9.14044155e-03,
            9.41496805e-03,
            -1.05606590e-15,
            2.05687455e-02,
            3.94621344e-17,
            -2.88662970e-16,
            1.05753983e-02,
            1.63964685e-17,
            2.49238075e-17,
            -4.81848153e-19,
            3.17051204e-04,
            2.62169259e-20,
            2.39682565e-03,
            -7.28733461e-03,
            -1.28765335e-03,
            -9.33859117e-04,
            9.51091881e-04,
            1.01855960e-18,
        ],
        [
            -1.82825857e-01,
            -6.21034004e-02,
            -1.95647158e-02,
            -3.00460554e-02,
            -9.43816211e-18,
            -3.89462688e-03,
            3.70148125e-01,
            6.79883311e-03,
            -7.82252075e-01,
            -6.03932040e-01,
            -1.25280585e-16,
            -2.23992386e-01,
            -2.47073170e-01,
            -1.10686987e-16,
            2.87746784e-02,
            4.92283106e-18,
            -1.23052572e-02,
            5.81118490e-18,
            -7.18687310e-03,
            1.09164496e00,
            1.39117837e-01,
            6.99027040e-02,
            3.61701941e-02,
            2.88089113e-18,
        ],
        [
            -5.44533618e-02,
            -9.61597179e-03,
            5.33913963e-04,
            -5.85934154e-03,
            3.53441787e-19,
            1.04793075e-03,
            -4.99514672e-03,
            -2.10149161e-02,
            -1.16846597e-01,
            -6.92258663e-02,
            4.12625278e-17,
            -3.35266497e-02,
            -3.21740739e-02,
            1.87714612e-17,
            3.76223276e-03,
            -1.80015543e-19,
            -1.03132924e-03,
            8.21606118e-19,
            -1.11830271e-03,
            1.25162373e-01,
            1.81259125e-02,
            7.04321033e-03,
            4.18891165e-03,
            2.75662313e-18,
        ],
        [
            1.78017202e-02,
            -8.46875523e-04,
            -3.58554111e-03,
            -8.51918455e-04,
            -4.58943798e-18,
            -1.57743806e-03,
            4.97714495e-02,
            8.12476996e-03,
            -3.33347898e-02,
            -5.41620151e-02,
            -1.58714343e-16,
            -9.63776413e-03,
            -2.14321739e-02,
            -8.25292932e-17,
            1.34537162e-03,
            2.85893174e-18,
            -1.02965293e-03,
            7.25745307e-19,
            -8.99943520e-04,
            7.42064752e-02,
            7.99132506e-03,
            5.98878248e-03,
            1.99712643e-03,
            -5.34975575e-18,
        ],
        [
            -3.04673008e-02,
            -6.31631081e-03,
            6.77374849e-04,
            -5.91985709e-04,
            -9.88185386e-19,
            -1.24259781e-03,
            3.88881985e-02,
            2.16317976e-02,
            -4.31553775e-02,
            1.46104139e-02,
            2.64290928e-17,
            -1.26425560e-02,
            9.36303961e-03,
            1.45772654e-17,
            1.38747837e-03,
            -5.07224125e-19,
            8.74642655e-05,
            3.16730519e-19,
            7.93313652e-04,
            3.91264596e-02,
            4.78786441e-03,
            2.01642557e-03,
            3.27243895e-03,
            1.44242955e-18,
        ],
        [
            -1.02684651e-17,
            -3.91183540e-18,
            9.42991911e-19,
            -2.28561005e-18,
            2.13249692e-03,
            8.72396666e-17,
            -7.96183700e-18,
            4.74068277e-18,
            -2.29038601e-17,
            -5.96652123e-18,
            5.87256071e-02,
            -3.70275263e-18,
            -2.07091148e-18,
            3.32108558e-02,
            1.60664064e-18,
            -1.56587251e-03,
            -2.97717004e-18,
            -4.92430952e-04,
            1.12261875e-18,
            2.86749844e-18,
            2.61930689e-18,
            -5.46383272e-18,
            1.44017430e-18,
            2.90409560e-03,
        ],
    ]

    return {
        "ref_pccd_dm1_no_fc": pccd_no_fc,
        "ref_pccd_dm1_fc": pccd_fc,
        "ref_lccsd_dm1_no_fc": lccsd_no_fc,
        "ref_lccsd_dm1_fc": lccsd_fc,
    }


@pytest.fixture(params=test_hf_dm)
def mol_generation(request, linalg):
    """Generate molecule for HF calculations and multipole moments"""
    mol = Molecule(linalg, "cc-pvdz", f"test/{request.param[0]}.xyz", alpha=0)
    mol.do_scf(RHF)
    mol.compute_dipole()

    # Return molecule and expected values
    return mol, request.param[1]


@pytest.fixture(params=test_opccd_dm)
def mol_generation_oopccd(request, linalg_slow):
    """Generate molecule for pCCD-based calculations"""
    mol = Molecule(
        linalg_slow, "cc-pvdz", f"test/{request.param[0]}.xyz", alpha=0
    )
    mol.do_scf(RHF)
    mol.do_pccd(request.param[1])
    mol.compute_dipole()

    # Return molecule and expected values
    return mol, request.param[2]


@pytest.fixture(params=ncore)
def mol_generation_oopccd_ncore(linalg_slow, request):
    """Generate molecule for pCCD-based calculations with frozen core orbitals"""
    mol = Molecule(
        linalg_slow, "cc-pvdz", "test/water.xyz", alpha=0, ncore=request.param
    )
    mol.do_scf(RHF)
    mol.do_pccd(ROOpCCD)
    # Return molecule
    return mol


############################>Quadrupole moment<############################
test_hf_quadrupole_moment = [
    (
        "water",
        {
            "quadrupole_hf": [
                -1.58341900,  # qm_xx
                0.09143664,  # qm_yy
                1.49198236,  # qm_zz
                0.00000000,  # qm_xy
                0.00000000,  # qm_xz
                0.00000000,  # qm_yz
            ],
        },
    ),
]

test_opccd_quadrupole_moment = [
    (
        "he",
        ROOpCCD,
        {
            "quadrupole_pccd": [
                0.00000000,
                0.00000000,
                0.00000000,
                0.00000000,
                0.00000000,
                0.00000000,
            ],
            "quadrupole_lccsd": [
                0.00000000,
                0.00000000,
                0.00000000,
                0.00000000,
                0.00000000,
                0.00000000,
            ],
        },
    ),
]


@pytest.fixture(params=test_hf_quadrupole_moment)
def mol_generation_quadrupole_moment(request, linalg):
    """Generate molecule for the HF quadrupole moment calculation"""
    mol = Molecule(linalg, "cc-pvdz", f"test/{request.param[0]}.xyz", alpha=0)
    mol.do_scf(RHF)
    mol.compute_quadrupole()

    # Return molecule and expected values
    return mol, request.param[1]


@pytest.fixture(params=test_opccd_quadrupole_moment)
def mol_generation_oopccd_qm(request, linalg):
    """Generate molecule for pCCD-based calculations"""
    mol = Molecule(linalg, "cc-pvdz", f"test/{request.param[0]}.xyz", alpha=0)
    mol.do_scf(RHF)
    mol.do_pccd(request.param[1])
    mol.compute_quadrupole()

    # Return molecule and expected values
    return mol, request.param[2]
